#include <fftw3.h>
#include <stdio.h>

//--- GSL Structures

struct gsl_block_struct
{
  size_t size;
  double *data;
};
typedef struct gsl_block_struct gsl_block;

typedef struct 
{
  size_t size;
  size_t stride;
  double *data;
  gsl_block *block;
  int owner;
} 
gsl_vector;

struct gsl_block_complex_struct
{
  size_t size;
  double *data;
};
typedef struct gsl_block_complex_struct gsl_block_complex;

typedef struct 
{
  size_t size;
  size_t stride;
  double *data;
  gsl_block_complex *block;
  int owner;
} gsl_vector_complex;


//--- Gambas-ish Structures

typedef
	struct
	{
		void*  GslVectorPtr;
		char   IsComplex;     // Gambas Boolean
	}
	GBP_Vector;  // Latter Part of Gambas CVECTOR


//============================================================================
//============================================================================
//    I N T E R N A L   R O U T I N E S
//============================================================================
//============================================================================



//============================================================================
//============================================================================
//    G A M B A S   E N T R Y   P O I N T S
//============================================================================
//============================================================================
char* FFT_GetVersion()
{
        return "Gambas FFTW Library Wrapper ver 0.0.1 by Cedron Dawg";
}
//============================================================================
char* FFT_RealForward( GBP_Vector* ArgSignalVector, GBP_Vector* ArgDftVector )
{
        int theSignalN, theDftN;

//--- Get the Signal Information
        
        gsl_vector* theSignalVec = ArgSignalVector->GslVectorPtr;
        
        double* theSignalArray = theSignalVec->data;
        
        theSignalN = theSignalVec->size;
        
//--- Get the Signal Information
        
        gsl_vector* theDftVec = ArgDftVector->GslVectorPtr;
        
        fftw_complex* theDftArray = (fftw_complex*) theDftVec->data;
        
        theDftN = theDftVec->size;

//--- Make Sure they are the Right size

        int theNeededSize;

        if( ( theSignalN & 1 ) > 0 )   // Odd
        {
            theNeededSize = ( theSignalN + 1 ) >> 1;
        }
        else
        {
            theNeededSize = ( theSignalN >> 1 ) + 1;
        }
        
        if( theDftN != theNeededSize ) return "Dft Size mismatch";

//--- Have FFTW do the DFT via a FFTW Plan

        fftw_plan thePlan;

        thePlan = fftw_plan_dft_r2c_1d( theSignalN, theSignalArray, theDftArray, FFTW_ESTIMATE);

        fftw_execute( thePlan );
        fftw_destroy_plan( thePlan );

//--- Apply 1/N Normalization to Output

        double theNormalizationFactor = 1.0 / (double) theSignalN;
        
        double* theValuePtr = (double*) theDftArray;
        
        int k;
        
        for( k = 0; k < theDftN; k++ )
        {
            *theValuePtr *= theNormalizationFactor; theValuePtr++;
            *theValuePtr *= theNormalizationFactor; theValuePtr++;
        }

//--- Exit

        return "";
}
//============================================================================
char* FFT_RealBackward( GBP_Vector* ArgSignalVector, GBP_Vector* ArgDftVector )
{
        int theSignalN;

//--- Get the Signal Information
        
        gsl_vector* theSignalVec = ArgSignalVector->GslVectorPtr;
        
        double* theSignalArray = theSignalVec->data;
        
        theSignalN = theSignalVec->size;
        
//--- Get the Signal Information
        
        gsl_vector* theDftVec = ArgDftVector->GslVectorPtr;
        
        fftw_complex* theDftArray = (fftw_complex*) theDftVec->data;
        
//--- Have FFTW do the DFT via a FFTW Plan

        fftw_plan thePlan;

        thePlan = fftw_plan_dft_c2r_1d( theSignalN, theDftArray, theSignalArray, FFTW_ESTIMATE);

        fftw_execute( thePlan );
        fftw_destroy_plan( thePlan );

//--- Exit

        return "";
}
//============================================================================
char* FFT_ComplexForward( GBP_Vector* ArgSignalVector, GBP_Vector* ArgDftVector )
{
        int theSignalN, theDftN;

//--- Get the Signal Information
        
        gsl_vector* theSignalVec = ArgSignalVector->GslVectorPtr;
        
        fftw_complex* theSignalArray = (fftw_complex*) theSignalVec->data;
        
        theSignalN = theSignalVec->size;
        
//--- Get the Signal Information
        
        gsl_vector* theDftVec = ArgDftVector->GslVectorPtr;
        
        fftw_complex* theDftArray = (fftw_complex*) theDftVec->data;
        
        theDftN = theDftVec->size;

//--- Make Sure they are the Right size

        if( theDftN != theSignalN ) return "Dft Size mismatch";

//--- Have FFTW do the DFT via a FFTW Plan

        fftw_plan thePlan;

        thePlan = fftw_plan_dft_1d( theSignalN, theSignalArray, theDftArray, FFTW_FORWARD, FFTW_ESTIMATE);

        fftw_execute( thePlan );
        fftw_destroy_plan( thePlan );

//--- Apply 1/N Normalization to Output

        double theNormalizationFactor = 1.0 / (double) theSignalN;
        
        double* theValuePtr = (double*) theDftArray;
        
        int k;
        
        for( k = 0; k < theDftN; k++ )
        {
            *theValuePtr *= theNormalizationFactor; theValuePtr++;
            *theValuePtr *= theNormalizationFactor; theValuePtr++;
        }

//--- Exit

        return "";
}
//============================================================================
char* FFT_ComplexBackward( GBP_Vector* ArgSignalVector, GBP_Vector* ArgDftVector )
{
        int theSignalN, theDftN;

//--- Get the Signal Information
        
        gsl_vector* theSignalVec = ArgSignalVector->GslVectorPtr;
        
        fftw_complex* theSignalArray = (fftw_complex*) theSignalVec->data;
        
        theSignalN = theSignalVec->size;
        
//--- Get the Signal Information
        
        gsl_vector* theDftVec = ArgDftVector->GslVectorPtr;
        
        fftw_complex* theDftArray = (fftw_complex*) theDftVec->data;
        
        theDftN = theDftVec->size;

//--- Make Sure they are the Right size

        if( theDftN != theSignalN ) return "Dft Size mismatch";

//--- Have FFTW do the DFT via a FFTW Plan

        fftw_plan thePlan;

        thePlan = fftw_plan_dft_1d( theSignalN, theDftArray, theSignalArray, FFTW_BACKWARD, FFTW_ESTIMATE);

        fftw_execute( thePlan );
        fftw_destroy_plan( thePlan );

//--- Exit

        return "";
}
//============================================================================
//============================================================================
//    H O U S E K E E P I N G
//============================================================================
//============================================================================
__attribute__((constructor)) void init(void)
{


}
//============================================================================
__attribute__((destructor)) void fini(void)
{

}
//============================================================================



